import uuid
from mmal.proto import (
    beta_create_MMAL_stub,
    beta_implementations,
    Path,
    PathRequest,
    PingRequest,
    Request,
    TimeSeriesRequest,
)


class Client(object):
    def __init__(self, host, port):
        self.channel = beta_implementations.insecure_channel(host, port)
        self.stub =  beta_create_MMAL_stub(self.channel)

    def request(self, request):
        return self.stub.RPCRequest(request, 10)

    def _mk_request(self):
        return Request(id=str(uuid.uuid1()))

    def _mk_paths(self, paths):
        return [
            Path(parts=path)
            for path in paths
        ]

    def ping_request(self, paths):
        ping_req = PingRequest()
        req = Request(ping_request=ping_req)
        req.id = str(uuid.uuid1())
        req.type = req.PING

        return self.request(req)

    def path_request(self, paths):
        path_req = PathRequest()
        path_req.paths.extend(self._mk_paths(paths))
        req = Request(path_request=path_req)
        req.id = str(uuid.uuid1())
        req.type = req.PATH

        return self.request(req)

    def ts_request(self, paths, cols=[], offset=0, limit=10):
        ts_req = TimeSeriesRequest(offset=offset, limit=limit)
        ts_req.paths.extend(self._mk_paths(paths))
        ts_req.columns.extend(list(set(cols)))
        req = Request(ts_request=ts_req)
        req.id = str(uuid.uuid1())
        req.type = req.TIMESERIES

        return self.request(req)



from __future__ import print_function
import os
import sys
from mmal.client import Client


def splitall(path):
    allparts = []
    while 1:
        parts = os.path.split(path)
        if parts[0] == path:  # sentinel for absolute paths
            allparts.insert(0, parts[0])
            break
        elif parts[1] == path: # sentinel for relative paths
            allparts.insert(0, parts[1])
            break
        else:
            path = parts[0]
            allparts.insert(0, parts[1])

    return allparts


def mmal_cmd(subparsers):
    path = subparsers.add_parser('path')
    path_subparser(path)

    ping = subparsers.add_parser('ping')
    ping_subparser(ping)

    ts = subparsers.add_parser('ts')
    ts_subparser(ts)


def path_subparser(subparser):
    subparser.set_defaults(func=path_req)
    subparser.add_argument(
        '-p', '--port',
        default = os.environ.get('MMAL_PORT', 8080),
        help    = 'Default port',
    )
    subparser.add_argument(
        '--host',
        default = os.environ.get('MMAL_HOST', 'localhost'),
        help    = 'Default host',
    )
    subparser.add_argument(
        'path',
        default = [],
	nargs   = '*',
        help    = 'path',
    )


def ping_subparser(subparser):
    subparser.set_defaults(func=ping_req)
    subparser.add_argument(
        '-p', '--port',
        default = os.environ.get('MMAL_PORT', 8080),
        help    = 'Default port',
    )
    subparser.add_argument(
        '--host',
        default = os.environ.get('MMAL_HOST', 'localhost'),
        help    = 'Default host',
    )
    subparser.add_argument(
        'path',
        default = [],
	nargs   = '*',
        help    = 'path',
    )


def ts_subparser(subparser):
    subparser.set_defaults(func=ts_req)
    subparser.add_argument(
        '-p', '--port',
        default = os.environ.get('MMAL_PORT', 8080),
        help    = 'Default port',
    )
    subparser.add_argument(
        '--host',
        default = os.environ.get('MMAL_HOST', 'localhost'),
        help    = 'Default host',
    )
    subparser.add_argument(
        '-c', '--columns',
        default = [],
        action  = 'append',
        help    = 'Columns',
    )
    subparser.add_argument(
        '-o', '--offset',
        default = 0,
        type    = int,
        help    = 'Offset',
    )
    subparser.add_argument(
        '-l', '--limit',
        default = 10,
        type    = int,
        help    = 'Limit',
    )
    subparser.add_argument(
        'path',
        default = [],
	nargs   = '*',
        help    = 'path',
    )


def path_req(args):
    if len(args.path) == 0:
        sys.exit(1)

    paths = [ splitall(x) for x in args.path ]
    client = Client(args.host, args.port)

    print(client.path_request(paths))


def ping_req(args):
    paths = [ splitall(x) for x in args.path ]

    client = Client(args.host, args.port)

    print(client.ping_request(paths))


def ts_req(args):
    paths = [ splitall(x) for x in args.path ]

    client = Client(args.host, args.port)

    reply = client.ts_request(
        paths,
        cols   = args.columns,
        offset = args.offset,
        limit  = args.limit,
    )
    print(reply)



from pkgutil import iter_modules
import argparse
import importlib
import os
import sys


def find_subcommands(subparsers):
    for _, mod_name, is_pkg in iter_modules([os.path.dirname(__file__)]):
        if not is_pkg and mod_name not in [sys.modules, 'cli']:
            mod_fullname = 'mmal.bin.{}'.format(mod_name)
            module = importlib.import_module(mod_fullname)
            if hasattr(module, 'mmal_cmd'):
                module.mmal_cmd(subparsers)


def main():
    parser = argparse.ArgumentParser(
        prog = "mmal",
        description = "mmal: Meteorological Middleware Application Layer",
    )

    subparsers = parser.add_subparsers()

    find_subcommands(subparsers)

    args = parser.parse_args()

    if hasattr(args, 'func'):
        args.func(args)



from mmal.proto import *
import sys


if sys.maxint > 2**32:
     INT_64 = True
else:
     INT_64 = False


def get_col(ts, index):
    col_type = ts.columns[index].type
    if col_type == TYPE_BOOL:
        return ts.columns[index].bools
    elif col_type == TYPE_BYTES:
        return ts.columns[index].raw_bytes
    elif col_type == TYPE_DOUBLE:
        return ts.columns[index].doubles
    elif col_type == TYPE_FIXED32:
        return ts.columns[index].fixed32s
    elif col_type == TYPE_FIXED64:
        return ts.columns[index].fixed64s
    elif col_type == TYPE_FLOAT:
        return ts.columns[index].floats
    elif col_type == TYPE_INT32:
        return ts.columns[index].int32s
    elif col_type == TYPE_INT64:
        return ts.columns[index].int64s
    elif col_type == TYPE_SFIXED32:
        return ts.columns[index].sfixed32s
    elif col_type == TYPE_SFIXED64:
        return ts.columns[index].sfixed64s
    elif col_type == TYPE_SINT32:
        return ts.columns[index].sint32s
    elif col_type == TYPE_SINT64:
        return ts.columns[index].sint64s
    elif col_type == TYPE_STRING:
        return ts.columns[index].strings
    elif col_type == TYPE_UINT32:
        return ts.columns[index].uint32s
    elif col_type == TYPE_UINT64:
        return ts.columns[index].uint64s

def extend_col(col, x):
    if isinstance(x, basestring):
        col.type = TYPE_STRING
        col.strings.extend([x])
    if isinstance(x, bool):
        col.type = TYPE_BOOL
        col.bools.extend([x])
    elif isinstance(x, bytes):
        col.type = TYPE_BYTES
        col.raw_bytes.extend([x])
    elif isinstance(x, bytearray):
        col.type = TYPE_BYTES
        col.raw_bytes.extend([x])
    elif isinstance(x, float):
        col.type = TYPE_FLOAT
        col.floats.extend([x])

    if INT_64:
        if isinstance(x, (int, long)):
            col.type = TYPE_INT64
            col.int64s.extend([x])
    else:
        if isinstance(x, (int, long)):
            col.type = TYPE_INT32
            col.int32s.extend([x])

def get_type(x):
    if isinstance(x, basestring):
        return TYPE_STRING
    if isinstance(x, bool):
        return TYPE_BOOL
    elif isinstance(x, bytes):
        return TYPE_BYTES
    elif isinstance(x, bytearray):
        return TYPE_BYTES
    elif isinstance(x, float):
        return TYPE_FLOAT

    if INT_64:
        if isinstance(x, (int, long)):
            return TYPE_INT64
    else:
        if isinstance(x, (int, long)):
            return TYPE_INT32



from mmal.server import run_server
from mmal.service import BaseMmalRPC, MmalRPCSqlAlchemy

def mmal_cmd(subparsers):
    server_subparser = subparsers.add_parser('server')
    server_subparser.set_defaults(func=main)
    server_subparser.add_argument(
        '-p', '--port',
        default = 8080,
        help    = 'Default port',
    )
    server_subparser.add_argument(
        '-d', '--dsn',
        default = '',
        help    = 'SQL dsn',
    )

def main(args):
    if len(args.dsn) > 0:
        rpc = MmalRPCSqlAlchemy()
        rpc.connect(args.dsn)
        run_server(args.port, rpc)
    else:
        run_server(args.port, BaseMmalRPC())




from sqlalchemy import *
from sqlalchemy.sql.expression import select, table as TABLE
from sqlalchemy.sql.expression import column as COLUMN
from mmal.utils import extend_col
from mmal.proto import (
    BetaMMALServicer,
    Reply,
    Path,
    TimeSeries,
)

def exception_handler(func):
    def inner(*args, **kwargs):
        try:
	    return func(*args, **kwargs)
        except Exception as e:
            req = args[1]
	    reply = Reply()

	    if req.type == req.PING:
		reply.type = reply.PONG
            elif req.type == req.PATH:
		reply.type = reply.PATH
            elif req.type == req.TIMESERIES:
		reply.type = reply.TIMESERIES

	    reply.error.type = reply.error.EXCEPTION
            reply.error.message = str(e)

            return reply

    return inner

class BaseMmalRPC(BetaMMALServicer):

    @exception_handler
    def RPCRequest(self, request, context):
        if request.type == request.PING:
            return self.on_ping(request)
        elif request.type == request.PATH:
            return self.on_path(request)
        elif request.type == request.TIMESERIES:
            return self.on_ts(request)

    def on_ping(self, request):
        reply = Reply()
        reply.id = request.id
        reply.type = reply.PONG
        reply.pong_reply.pong = "PONG"
        return reply

    def on_path(self, request):
        reply = Reply()
        reply.id = request.id
        reply.type = reply.PATH
        return reply

    def on_ts(self, request):
        reply = Reply()
        reply.id = request.id
        reply.type = reply.TIMESERIES
        return reply

def strip_path(path):
    if path[0] == u'/':
        return path[1:]

    return path

class MmalRPCSqlAlchemy(BaseMmalRPC):
    sql_engine = None
    tables = dict()

    def connect(self, dsn):
        self.sql_engine = create_engine(dsn)
        inspector = inspect(self.sql_engine)
        meta = MetaData()
        meta.reflect(bind=self.sql_engine)
        self.meta = meta
        tables = inspector.get_table_names()

        for table in tables:
            self.tables[table] = inspector.get_columns(table)

    def on_path(self, request):
        reply = super(MmalRPCSqlAlchemy, self).on_path(request)
        req_paths = [ x.parts for x in request.path_request.paths[:] ]

        if any([ x[0] == u'/' and len(x) <= 1 for x in req_paths ]):
            for table, cols in self.tables.items():
                path = Path()
                path.type = path.LEAF
                path.parts.extend([table])
                path.columns.extend([ x.get('name', '') for x in cols])
                reply.path_reply.paths.extend([path])

        else:
            for req_path in req_paths:
                if req_path[-1] not in self.tables:
                    reply.error.type = reply.error.INVALID_PATH
                    reply.error.message = "no such path {}".format(req_path)
                    break

                cols = self.tables[req_path[-1]]
                path = Path()
                path.type = path.LEAF
                path.parts.extend(strip_path(req_path))
                path.columns.extend([ x.get('name', '') for x in cols])
                reply.path_reply.paths.extend([path])

        return reply

    def on_ts(self, request):
        reply = super(MmalRPCSqlAlchemy, self).on_ts(request)
        req_paths = [ x.parts for x in request.ts_request.paths[:] ]
        req_cols = request.ts_request.columns[:]

        if any([ x[0] == u'/' and len(x) <= 1 for x in req_paths ]) or len(req_paths) == 0:
            for table, cols in self.tables.items():
                col_names = [ x.get('name', '') for x in cols]

                if len(req_cols) == 0:
                    req_cols = col_names

                ts = TimeSeries()
                if len(req_cols) > 0:
                    ts.columns.extend([
                        ts.Column(name=x)
                        for x in req_cols if x in col_names
                    ])
                else:
                    ts.columns.extend([
                        ts.Column(name=x)
                        for x in col_names
                    ])

                with self.sql_engine.connect() as conn:
                    sql_table = self.meta.tables[table]
                    statement = select(
                        from_obj=sql_table,
                    ).with_only_columns([
                        sql_table.columns.get(str(x))
                        for x in req_cols if str(x) in sql_table.columns
                    ])

                    statement = statement.limit(request.ts_request.limit)
                    statement = statement.offset(request.ts_request.offset)

                    rows = conn.execute(statement)

                    for row in rows:
                        row_idx = 0
                        for i, col_data in enumerate(row):
                            if col_data is not None and len(ts.columns)-1 >= row_idx:
                                extend_col(ts.columns[row_idx], col_data)
                                row_idx += 1

                # set path info
                path = Path()
                path.type = path.LEAF
                path.parts.extend([table])
                path.columns.extend(col_names)
                reply.ts_reply.time_series.extend([ts])
                reply.ts_reply.paths.extend([path])

        else:
            for req_path in req_paths:
                if req_path[-1] not in self.tables:
                    reply.error.type = reply.error.INVALID_PATH
                    reply.error.message = "no such path {}".format(req_path)
                    break

                table = req_path[-1]
                cols = self.tables[req_path[-1]]
                col_names = [ x.get('name', '') for x in cols]
                if len(req_cols) == 0:
                    req_cols = col_names

                ts = TimeSeries()
                if len(req_cols) > 0:
                    ts.columns.extend([
                        ts.Column(name=x)
                        for x in req_cols if x in col_names
                    ])
                else:
                    ts.columns.extend([
                        ts.Column(name=x)
                        for x in col_names
                    ])

                with self.sql_engine.connect() as conn:
                    sql_table = self.meta.tables[table]
                    statement = select(
                        from_obj=sql_table,
                    ).with_only_columns([
                        sql_table.columns.get(str(x))
                        for x in req_cols if str(x) in sql_table.columns
                    ])
                    statement = statement.limit(request.ts_request.limit)
                    statement = statement.offset(request.ts_request.offset)

                    rows = conn.execute(statement)

                    for row in rows:
                        row_idx = 0
                        for i, col_data in enumerate(row):
                            if col_data is not None and len(ts.columns)-1 >= row_idx:
                                extend_col(ts.columns[row_idx], col_data)
                                row_idx += 1

                    # set path info
                    path = Path()
                    path.type = path.LEAF
                    path.parts.extend([table])
                    path.columns.extend(col_names)
                    reply.ts_reply.time_series.extend([ts])
                    reply.ts_reply.paths.extend([path])

        return reply



from setuptools import setup, find_packages

setup(
    name             = "mmal",
    description      = "Meteorological Middleware Application Layer",
    url              = "https://github.com/hodgesds/mmal-python",
    version          = "0.0.2",
    author           = "Daniel Hodges",
    author_email     = "hodges.daniel.scott@gmail.com",
    scripts          = [ "bin/mmal" ],
    install_requires = [ "mmal-proto", "grpcio", "sqlalchemy" ],
    test_suite       = "",
    tests_require    = [ "tox", "nose" ],
    packages         = find_packages(
        where        = '.',
        exclude      = ('tests*', 'bin*'),
    ),
)



from mmal.proto import beta_create_MMAL_server

def run_server(port, rpc):
    server = beta_create_MMAL_server(rpc)
    server.add_insecure_port('[::]:{}'.format(port))
    server.start()
    try:
	while True:
	    pass
    except KeyboardInterrupt:
        server.stop()
